#!/usr/bin/python3
"""
// © University of Southampton IT Innovation Centre, 2018
//
// Copyright in this software belongs to University of Southampton
// IT Innovation Centre of Gamma House, Enterprise Road,
// Chilworth Science Park, Southampton, SO16 7NS, UK.
//
// This software may not be used, sold, licensed, transferred, copied
// or reproduced in whole or in part in any manner or form or in or
// on any media by any person other than in accordance with the terms
// of the Licence Agreement supplied with the software, or otherwise
// without the prior written consent of the copyright owners.
//
// This software is distributed WITHOUT ANY WARRANTY, without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE, except where stated in the Licence Agreement supplied with
// the software.
//
//      Created By :            Nikolay Stanchev
//      Created Date :          14-08-2018
//      Created for Project :   FLAME
"""

# Python standard libs
import logging
from hashlib import sha256
from collections import defaultdict

# PIP installed libs
from pyramid.httpexceptions import HTTPBadRequest
from pyramid.view import view_defaults, view_config
from yaml import load, YAMLError
from toscaparser.tosca_template import ToscaTemplate
from requests import post, get, delete

# CLMC-service imports
from clmcservice.alertsapi.utilities import adjust_tosca_definitions_import, TICKScriptTemplateFiller, fill_http_post_handler_vars, get_resource_spec_policy_triggers, get_alert_spec_policy_triggers
from clmcservice.alertsapi.alerts_specification_schema import COMPARISON_OPERATORS, SFEMC, validate_clmc_alerts_specification

# initialise logger
log = logging.getLogger('service_logger')


@view_defaults(renderer='json')
class AlertsConfigurationAPI(object):
    """
    A class-based view for configuring alerts within CLMC.
    """

    STREAM_PERIOD_LIMIT = 60  # if alert period is <= 60 seconds, then a stream template is used, otherwise use batch

    DUAL_VERSION_TEMPLATES = {"threshold"}  # this set defines all template types that are written in two versions (stream and batch)

    KAPACITOR_TASK_API_PREFIX = "/kapacitor/v1/tasks"

    KAPACITOR_TOPIC_API_PREFIX = "/kapacitor/v1/alerts/topics"

    def __init__(self, request):
        """
        Initialises the instance of the view with the request argument.

        :param request: client's call request
        """

        self.request = request

    @view_config(route_name='alerts_configuration', request_method='GET')
    def get_alerts_hash(self):
        """
        (DEPRECATED - there is a GET /alerts/<sfc>/<sfc_instance> API method)
        Retrieves hash value for alerts task, topic and handlers based on sfc, sfci, policy and trigger IDs
        """

        for param in ("sfc", "sfci", "policy", "trigger"):
            if param not in self.request.params:
                raise HTTPBadRequest("Missing url query string parameter {0}.".format(param))

        sfc = self.request.params["sfc"]
        sfci = self.request.params["sfci"]
        policy = self.request.params["policy"]
        trigger = self.request.params["trigger"]

        # generate topic and alert identifiers
        topic_id = "{0}\n{1}\n{2}\n{3}".format(sfc, sfci, policy, trigger)
        topic_id = self.get_hash(topic_id)
        alert_id = topic_id

        return {
            "task_identifier": alert_id,
            "task_api_endpoint": "{0}/{1}".format(self.KAPACITOR_TASK_API_PREFIX, alert_id),
            "topic_identifier": topic_id,
            "topic_api_endpoint": "{0}/{1}".format(self.KAPACITOR_TOPIC_API_PREFIX, topic_id),
            "topic_handlers_api_endpoint": "{0}/{1}/handlers".format(self.KAPACITOR_TOPIC_API_PREFIX, topic_id)
        }

    @view_config(route_name='alerts_configuration_aggregate', request_method='GET')
    def get_all_sfc_alerts(self):
        """
        The view for retrieving all alerts and Kapacitor resources registered for instances of a specific service function chain.

        :return: a mapping between SFC instance IDs and list of alert objects,
            each object representing an alert and containing the policy ID, trigger ID, task ID, topic ID and Kapacitor endpoints for these
        """

        # fetch the URL parameter
        sfc_id = self.request.matchdict["sfc_id"]

        return self._get_all_alerts(sfc_id)

    @view_config(route_name='alerts_configuration_instance', request_method='GET')
    def get_all_sfc_instance_alerts(self):
        """
        The view for retrieving all alerts and Kapacitor resources registered for a specific service function chain instance.

        :return: a list of alert objects, each object representing an alert and containing the policy ID, trigger ID, task ID, topic ID and kapacitor endpoints for these
        """

        # fetch the URL parameters
        sfc_id, sfc_instance_id = self.request.matchdict["sfc_id"], self.request.matchdict["sfc_instance_id"]

        # return the list of alert objects for the SFC instance parameter
        return self._get_all_alerts(sfc_id, sfc_instance_id).get(sfc_instance_id, [])  # defaults to empty list if no alerts are found for this SFC instance

    def _get_all_alerts(self, sfc_id, sfc_instance_id=None):
        """
        Fetch all alerts configured for a specific SFC or SFC instance.

        :param sfc_id: the SFC identifier
        :param sfc_instance_id: the SFC instance identifier (defaults to None), if set to None, return the alerts for all instances of this SFC

        :return: metadata of the CLMC alerts that were configured for this SFC (a mapping between SFC instance IDs and list of alert objects)
        """

        kapacitor_host, kapacitor_port = self.request.registry.settings['kapacitor_host'], self.request.registry.settings['kapacitor_port']
        sfemc_fqdn, sfemc_port = self.request.registry.settings['sfemc_fqdn'], self.request.registry.settings['sfemc_port']
        fqdn_prefix = "http://{0}:{1}/sfemc/event".format(sfemc_fqdn, sfemc_port)

        # fetch all alerts that are configured for this SFC
        tasks_to_return = defaultdict(list)

        # traverse every registered task and check if it is configured for this SFC instance
        for task in self._fetch_alert_tasks_for_sfc(sfc_id, sfc_instance_id):

            # get the configured variables of this alert
            task_config = task["vars"]

            # get the TOSCA spec. configuration values
            task_id = task["id"]
            topic_id = task_id
            policy_id = task_config["policy"]["value"]
            trigger_id = task_config["eventID"]["value"]

            kapacitor_handlers_relative_url = "{0}/{1}/handlers".format(self.KAPACITOR_TOPIC_API_PREFIX, topic_id)
            kapacitor_handlers_full_url = "http://{0}:{1}{2}".format(kapacitor_host, kapacitor_port, kapacitor_handlers_relative_url)
            handlers = [handler_obj["options"]["url"] for handler_obj in get(kapacitor_handlers_full_url).json()["handlers"]]
            handlers = list(map(lambda handler: SFEMC if handler.startswith(fqdn_prefix) else handler, handlers))

            alert_object = {"policy": policy_id, "trigger": trigger_id, "handlers": handlers,
                            "task_identifier": task_id, "topic_identifier": topic_id,
                            "task_api_endpoint": "{0}/{1}".format(self.KAPACITOR_TASK_API_PREFIX, task_id),
                            "topic_api_endpoint": "{0}/{1}".format(self.KAPACITOR_TOPIC_API_PREFIX, topic_id),
                            "topic_handlers_api_endpoint": kapacitor_handlers_relative_url}

            # add it to the list of alerts for this SFC instance
            tasks_to_return[task_config["sfci"]["value"]].append(alert_object)

        return tasks_to_return

    @view_config(route_name='alerts_configuration_aggregate', request_method='DELETE')
    def delete_all_sfc_alerts(self):
        """
        The view function for deleting all alerts and Kapacitor resources registered for all instances of a specific service function chain.

        :return: the number of deleted alerts for each SFC instance.
        """

        # fetch the URL parameter
        sfc_id = self.request.matchdict["sfc_id"]

        return self._delete_all_alerts(sfc_id)

    @view_config(route_name='alerts_configuration_instance', request_method='DELETE')
    def delete_all_sfc_instance_alerts(self):
        """
        The view function for deleting all alerts and Kapacitor resources registered for a specific service function chain instance.

        :return: the number of deleted alerts for this specific SFC instance.
        """

        # fetch the URL parameters
        sfc_id, sfc_instance_id = self.request.matchdict["sfc_id"], self.request.matchdict["sfc_instance_id"]

        # if no SFC instance alerts were deleted, manually set the count to 0
        return self._delete_all_alerts(sfc_id, sfc_instance_id).get(sfc_instance_id, {"deleted_alerts_count": 0})  # defaults to 0 deleted alerts if SFC instance is not found

    def _delete_all_alerts(self, sfc_id, sfc_instance_id=None):
        """
        Delete all alerts configured for a specific SFC or SFC instance.

        :param sfc_id: the SFC identifier
        :param sfc_instance_id: the SFC instance identifier (defaults to None), if set to None, delete the alerts for all instances of this SFC

        :return: a mapping between SFC instance IDs and dictionary showing the count of deleted alerts)
        """

        kapacitor_host, kapacitor_port = self.request.registry.settings['kapacitor_host'], self.request.registry.settings['kapacitor_port']

        deleted_count_dict = defaultdict(dict)

        # traverse every registered task and check if it is configured for this SFC instance
        for task in self._fetch_alert_tasks_for_sfc(sfc_id, sfc_instance_id):

            # delete the Kapacitor task
            task_id = task["id"]
            kapacitor_task_relative_url = "{0}/{1}".format(self.KAPACITOR_TASK_API_PREFIX, task_id)
            kapacitor_task_full_url = "http://{0}:{1}{2}".format(kapacitor_host, kapacitor_port, kapacitor_task_relative_url)
            log.info("Sending DELETE request to Kapacitor for URL - {0}".format(kapacitor_task_full_url))
            response = delete(kapacitor_task_full_url)
            log.info("Received status code {0} from Kapacitor for URL - {1}".format(response.status_code, kapacitor_task_full_url))

            # log the response body if the status code was different than 204
            assert response.status_code == 204, "Received response status code different than 204 when deleting an alert task - {0}.".format(response.text)

            # increment the count of deleted alerts
            delete_dict = deleted_count_dict[task["vars"]["sfci"]["value"]]
            try:
                delete_dict["deleted_alerts_count"] += 1
            except KeyError:
                delete_dict["deleted_alerts_count"] = 1

            # get the Kapacitor topic and the subscribed handlers
            topic_id = task_id
            kapacitor_topic_relative_url = "{0}/{1}".format(self.KAPACITOR_TOPIC_API_PREFIX, topic_id)
            kapacitor_topic_full_url = "http://{0}:{1}{2}".format(kapacitor_host, kapacitor_port, kapacitor_topic_relative_url)
            kapacitor_topic_handlers_full_url = "{0}{1}".format(kapacitor_topic_full_url, "/handlers")

            log.info("Sending GET request to Kapacitor for URL - {0}".format(kapacitor_topic_handlers_full_url))
            response = get(kapacitor_topic_handlers_full_url)
            log.info("Received status code {0} from Kapacitor for URL - {1}".format(response.status_code, kapacitor_topic_handlers_full_url))

            # traverse each handler and delete it
            for handler in response.json()["handlers"]:
                handler_id = handler["id"]
                kapacitor_topic_handlers_instance_full_url = "{0}/{1}".format(kapacitor_topic_handlers_full_url, handler_id)
                log.info("Sending DELETE request to Kapacitor for URL - {0}".format(kapacitor_topic_handlers_instance_full_url))
                response = delete(kapacitor_topic_handlers_instance_full_url)
                log.info("Received status code {0} from Kapacitor for URL - {1}".format(response.status_code, kapacitor_topic_handlers_instance_full_url))
                assert response.status_code == 204, "Received response status code different than 204 when deleting an alert topic handler - {0}.".format(response.text)

            # delete the Kapacitor topic
            log.info("Sending DELETE request to Kapacitor for URL - {0}".format(kapacitor_topic_full_url))
            response = delete(kapacitor_topic_full_url)
            log.info("Received status code {0} from Kapacitor for URL - {1}".format(response.status_code, kapacitor_topic_full_url))
            assert response.status_code == 204, "Received response status code different than 204 when deleting an alert topic - {0}.".format(response.text)

        return deleted_count_dict

    def _fetch_alert_tasks_for_sfc(self, sfc_id, sfc_instance_id=None):
        """
        Utility method to retrieve the alert tasks configured for a specific service function chain and service function chain instance.

        :param sfc_id: the SFC identifier
        :param sfc_instance_id: the SFC instance identifier (defaults to None), if set to None, return alerts for all instances of this SFC

        :return: the alerts configuration for the SFC identified with the function parameters, yielding each of these using an iterator pattern
        """

        kapacitor_host, kapacitor_port = self.request.registry.settings['kapacitor_host'], self.request.registry.settings['kapacitor_port']

        # get all tasks from Kapacitor
        kapacitor_tasks_url = "http://{0}:{1}{2}".format(kapacitor_host, kapacitor_port, self.KAPACITOR_TASK_API_PREFIX)
        log.info("Sending GET request to Kapacitor for URL - {0}".format(kapacitor_tasks_url))
        response = get(kapacitor_tasks_url)
        log.info("Received status code {0} from Kapacitor for URL - {1}".format(response.status_code, kapacitor_tasks_url))

        all_tasks = response.json()["tasks"]

        # traverse every registered task and check if it is configured for this SFC instance
        for task in all_tasks:

            # get the configured variables of this alert
            task_config = task["vars"]

            # if the SFC parameter matches
            if task_config["sfc"]["value"] == sfc_id:
                # if the SFC instance parameter matches or it was set to None
                if sfc_instance_id is None or task_config["sfci"]["value"] == sfc_instance_id:
                    yield task

    @view_config(route_name='alerts_configuration', request_method='DELETE')
    def delete_alerts_specification(self):
        """
        The view for deleting alerts based on a TOSCA alerts specification document.

        :return: a dictionary with the deleted alerts and handlers

        :raises HTTPBadRequest: if the request doesn't contain a (YAML) file input referenced as alert-spec representing the TOSCA Alerts Specification
        """

        kapacitor_host, kapacitor_port = self.request.registry.settings['kapacitor_host'], self.request.registry.settings['kapacitor_port']
        sfemc_fqdn, sfemc_port = self.request.registry.settings['sfemc_fqdn'], self.request.registry.settings['sfemc_port']
        fqdn_prefix = "http://{0}:{1}/sfemc/event".format(sfemc_fqdn, sfemc_port)

        # since this is not a POST request, need to fetch the alert specification file from request.params, rather than request.POST
        alert_spec_reference = self.request.params.get('alert-spec')

        # parse the alert specification file into a tosca template (including validation)
        tosca_tpl = self._parse_alert_spec(alert_spec_reference)

        # sfc, sfc_instance = tosca_tpl.tpl["metadata"]["sfc"], tosca_tpl.tpl["metadata"]["sfci"]
        sfc = tosca_tpl.tpl["metadata"]["servicefunctionchain"]
        sfc_instance = "{0}_1".format(sfc)

        alerts = []
        handlers = []

        for policy in tosca_tpl.policies:
            for trigger in policy.triggers:
                event_id = trigger.name
                policy_id = policy.name

                # generate topic and alert identifiers
                topic_id = "{0}\n{1}\n{2}\n{3}".format(sfc, sfc_instance, policy_id, event_id)  # scoped per service function chain instance (no two sfc instances report to the same topic)
                topic_id = self.get_hash(topic_id)
                task_id = topic_id

                # delete alert task
                kapacitor_api_task_url = "http://{0}:{1}{2}/{3}".format(kapacitor_host, kapacitor_port, self.KAPACITOR_TASK_API_PREFIX, task_id)
                if get(kapacitor_api_task_url).status_code == 200:
                    delete(kapacitor_api_task_url)
                    alerts.append({"policy": policy_id, "trigger": event_id})

                # get all alert handlers
                kapacitor_api_topic_handlers_url = "http://{0}:{1}{2}/{3}/handlers".format(kapacitor_host, kapacitor_port, self.KAPACITOR_TOPIC_API_PREFIX, topic_id)
                http_response = get(kapacitor_api_topic_handlers_url)
                if http_response.status_code != 200:
                    continue  # if the topic doesn't exist continue with the other triggers

                # delete alert handlers
                http_handlers = http_response.json()['handlers']
                for handler in http_handlers:

                    original_http_handler_url = handler["options"]["url"]
                    if original_http_handler_url.startswith(fqdn_prefix):
                        original_http_handler_url = SFEMC

                    handler_id = handler["id"]

                    kapacitor_api_handler_url = "http://{0}:{1}{2}/{3}/handlers/{4}".format(kapacitor_host, kapacitor_port, self.KAPACITOR_TOPIC_API_PREFIX, topic_id, handler_id)
                    delete(kapacitor_api_handler_url)
                    handlers.append({"policy": policy_id, "trigger": event_id, "handler": original_http_handler_url})

                # delete alert topic
                kapacitor_api_topic_url = "http://{0}:{1}{2}/{3}".format(kapacitor_host, kapacitor_port, self.KAPACITOR_TOPIC_API_PREFIX, topic_id)
                delete(kapacitor_api_topic_url)

        return {"deleted_alerts": alerts, "deleted_handlers": handlers}

    @view_config(route_name='alerts_configuration', request_method='PUT')
    def put_alerts_specification(self):
        """
        The view for receiving and configuring alerts based on the TOSCA alerts specification document. This endpoint must also receive the TOSCA resources specification document for validation.
        A PUT request will update any existing alerts and handlers that were registered before.

        :return: a dictionary with a msg and optional keys for errors encountered while interacting with Kapacitor

        :raises HTTPBadRequest: if the request doesn't contain a (YAML) file input referenced as alert-spec representing the TOSCA Alerts Specification
        :raises HTTPBadRequest: if the request doesn't contain a (YAML) file input referenced as resource-spec representing the TOSCA Resources Specification
        """

        return self._create_alerts_specification(patch_duplicates=True)

    @view_config(route_name='alerts_configuration', request_method='POST')
    def post_alerts_specification(self):
        """
        The view for receiving and configuring alerts based on the TOSCA alerts specification document. This endpoint must also receive the TOSCA resources specification document for validation.
        A POST request will not try to update existing alerts and handlers and will simply return the error message from Kapacitor.

        :return: a dictionary with a msg and optional keys for errors encountered while interacting with Kapacitor

        :raises HTTPBadRequest: if the request doesn't contain a (YAML) file input referenced as alert-spec representing the TOSCA Alerts Specification
        :raises HTTPBadRequest: if the request doesn't contain a (YAML) file input referenced as resource-spec representing the TOSCA Resources Specification
        """

        return self._create_alerts_specification(patch_duplicates=False)

    def _create_alerts_specification(self, patch_duplicates=False):
        """
        The base source code for creating (with or without updates) alerts from an alert specification document.

        :param patch_duplicates: (defaults to False) if set to True, existing resources will be overwritten

        :return: a dictionary with a msg and optional keys for errors encountered while interacting with Kapacitor
        """

        kapacitor_host, kapacitor_port = self.request.registry.settings['kapacitor_host'], self.request.registry.settings['kapacitor_port']

        if patch_duplicates:  # implying a PUT request
            alert_spec_reference = self.request.params.get('alert-spec')
            resource_spec_reference = self.request.params.get('resource-spec')
        else:  # implying a POST request
            alert_spec_reference = self.request.POST.get('alert-spec')
            resource_spec_reference = self.request.POST.get('resource-spec')

        # parse the resource specification file and extract the required information
        resource_spec_sfc, resource_spec_sfc_instance, resource_spec_policy_triggers = self._parse_resource_spec(resource_spec_reference)

        # parse the alert specification file into a tosca template (including validation)
        tosca_tpl = self._parse_alert_spec(alert_spec_reference)

        # extract the information needed from the alert spec. for validation between the two TOSCA documents
        alert_spec_policy_triggers = get_alert_spec_policy_triggers(tosca_tpl)

        # sfc, sfc_instance = tosca_tpl.tpl["metadata"]["sfc"], tosca_tpl.tpl["metadata"]["sfci"]
        sfc = tosca_tpl.tpl["metadata"]["servicefunctionchain"]
        sfc_instance = "{0}_1".format(sfc)

        # do validation between the two TOSCA documents
        self._compare_alert_and_resource_spec(sfc, sfc_instance, alert_spec_policy_triggers, resource_spec_sfc, resource_spec_sfc_instance, resource_spec_policy_triggers)

        db = sfc  # database per service function chain, named after the service function chain ID

        # iterate through every policy and extract all triggers of the given policy - the returned lists of errors will be empty if no errors were encountered
        # while interacting with the Kapacitor HTTP API, the patch_flag is set to False so that existing resources are not recreated
        alert_tasks_errors, alert_handlers_errors = self._config_kapacitor_alerts(tosca_tpl, sfc, sfc_instance, db,
                                                                                  kapacitor_host, kapacitor_port, resource_spec_policy_triggers,
                                                                                  patch_duplicates=patch_duplicates)

        return_msg = {"msg": "Alerts specification has been successfully validated and forwarded to Kapacitor", "service_function_chain_id": sfc,
                      "service_function_chain_instance_id": sfc_instance}

        if len(alert_tasks_errors) > 0:
            return_msg["triggers_specification_errors"] = alert_tasks_errors
            self.request.response.status = 400

        if len(alert_handlers_errors) > 0:
            return_msg["triggers_action_errors"] = alert_handlers_errors
            self.request.response.status = 400

        return return_msg

    def _compare_alert_and_resource_spec(self, alert_spec_sfc, alert_spec_sfc_instance, alert_spec_policy_triggers, resource_spec_sfc, resource_spec_sfc_instance, resource_spec_policy_triggers):
        """
        Compares the extracted values from the resource spec against the values from the alerts spec - validation that they refer to the same things,

        :param alert_spec_sfc: sfc from alert spec
        :param alert_spec_sfc_instance: sfc instance from alert spec
        :param alert_spec_policy_triggers: policy/trigger IDs from alert spec
        :param resource_spec_sfc: sfc from resource spec
        :param resource_spec_sfc_instance: sfc instance from resource spec
        :param resource_spec_policy_triggers: policy/trigger IDs from resource spec

        :raises: HTTP Bad Request if the two specifications are inconsistent
        """

        if alert_spec_sfc != resource_spec_sfc:
            raise HTTPBadRequest("Different service function chain ID used in the alert and resource specification documents: {0} != {1}".format(alert_spec_sfc, resource_spec_sfc))

        if alert_spec_sfc_instance != resource_spec_sfc_instance:
            raise HTTPBadRequest("Different service function chain instance ID used in the alert and resource specification documents: {0} != {1}".format(alert_spec_sfc_instance, resource_spec_sfc_instance))

        alert_spec_triggers_set = set(alert_spec_policy_triggers)
        resource_spec_triggers_set = set(resource_spec_policy_triggers)

        # check for triggers defined in the resource specification but missing in the alerts specification
        missing_policy_triggers = [policy_trigger_id.replace("\n", " : ") for policy_trigger_id in resource_spec_policy_triggers if policy_trigger_id not in alert_spec_triggers_set]
        if len(missing_policy_triggers) > 0:
            raise HTTPBadRequest("Couldn't match the following policy triggers from the resource specification with triggers defined in the alerts specification: {0}".format(missing_policy_triggers))

        # check for triggers defined in the alerts specification and posting to flame_sfemc, but missing in the resource specification
        missing_policy_triggers = [policy_trigger_id.replace("\n", " : ") for policy_trigger_id in alert_spec_policy_triggers if policy_trigger_id not in resource_spec_triggers_set]
        if len(missing_policy_triggers) > 0:
            raise HTTPBadRequest("Couldn't match the following policy triggers from the alerts specification with triggers defined in the resource specification: {0}".format(missing_policy_triggers))

    def _config_kapacitor_alerts(self, tosca_tpl, sfc, sfc_instance, db, kapacitor_host, kapacitor_port, resource_spec_policy_triggers, patch_duplicates=False):
        """
        Configures the alerts task and alert handlers within Kapacitor.

        :param tosca_tpl: the parsed Tosca template object
        :param sfc: sfc ID
        :param sfc_instance: sfc instance ID
        :param db: Influx database ID
        :param kapacitor_host: default host is localhost (CLMC service running on the same machine as Kapacitor)
        :param kapacitor_port: default value to use is 9092
        :param resource_spec_policy_triggers: the extracted policy-trigger strings from the resource specification
        :param patch_duplicates: (defaults to False) if set to True, any duplication errors will be handled by first deleting the existing resource and then creating it

        :return: the list for tracking errors while interacting with Kapacitor tasks and the list for tracking errors while interacting with Kapacitor alert handlers
        """

        kapacitor_api_tasks_url = "http://{0}:{1}{2}".format(kapacitor_host, kapacitor_port, self.KAPACITOR_TASK_API_PREFIX)

        # two lists to keep track of any errors while interacting with the Kapacitor HTTP API
        alert_tasks_errors = []
        alert_handlers_errors = []

        for policy in tosca_tpl.policies:
            for trigger in policy.triggers:
                event_id = trigger.name
                policy_id = policy.name
                resource_spec_trigger_id = resource_spec_policy_triggers.get("{0}\n{1}".format(policy_id, event_id))

                event_type = trigger.trigger_tpl["event_type"]
                template_id = "{0}-template".format(event_type)
                measurement, field = trigger.trigger_tpl["metric"].split(".")

                condition = trigger.trigger_tpl["condition"]
                critical_value = float(condition["threshold"])
                alert_period_integer = condition["granularity"]
                alert_period = "{0}s".format(alert_period_integer)
                influx_function = condition.get("aggregation_method", "mean")  # if not specified, use "mean"

                # check for tag filtering
                where_clause = None
                if "resource_type" in trigger.trigger_tpl["condition"]:
                    tags = condition["resource_type"]
                    # make sure alert tasks are executing with queries for the given sfc and sfc instance - automatically add those tags using the metadata values
                    tags["flame_sfc"] = sfc
                    tags["flame_sfci"] = sfc_instance

                    # build up the where clause from the tags dictionary
                    where_filters_list = map(lambda tag_name: '"{0}"=\'{1}\''.format(tag_name, tags[tag_name]), tags)
                    where_clause = " AND ".join(where_filters_list)

                comparison_operator = COMPARISON_OPERATORS[condition.get("comparison_operator", "gte")]  # if not specified, use "gte" (>=)

                # generate topic and alert identifiers
                topic_id = "{0}\n{1}\n{2}\n{3}".format(sfc, sfc_instance, policy_id, event_id)  # scoped per service function chain instance (no two sfc instances report to the same topic)
                topic_id = self.get_hash(topic_id)
                task_id = topic_id

                # check whether the template needs to be a stream or a batch
                if event_type in self.DUAL_VERSION_TEMPLATES:
                    if alert_period_integer < self.STREAM_PERIOD_LIMIT:
                        template_id = "{0}-stream-template".format(event_type)
                        event_type = "{0}_stream".format(event_type)
                    else:
                        template_id = "{0}-batch-template".format(event_type)
                        event_type = "{0}_batch".format(event_type)

                # built up the template vars dictionary depending on the event type (threshold, relative, etc.)
                # all extracted properties from the trigger are passed, the TICKScriptTemplateFiller entry point then forwards those to the appropriate function for template filling
                template_vars = TICKScriptTemplateFiller.fill_template_vars(event_type, sfc=sfc, sfci=sfc_instance, policy=policy_id, db=db, measurement=measurement,
                                                                            field=field, influx_function=influx_function, critical_value=critical_value, comparison_operator=comparison_operator,
                                                                            alert_period=alert_period, topic_id=topic_id, event_id=event_id, where_clause=where_clause)

                # create and activate alert task through the kapacitor HTTP API
                kapacitor_http_request_body = {
                    "id": task_id,
                    "template-id": template_id,
                    "dbrps": [{"db": db, "rp": "autogen"}],
                    "status": "enabled",
                    "vars": template_vars
                }

                # send the request and receive a response
                response = post(kapacitor_api_tasks_url, json=kapacitor_http_request_body)
                response_content = response.json()
                # log the response
                log.info("{0} - {1}".format(response_content, response.status_code))

                # track all reported errors
                if response_content.get("error", "") != "":  # Kapacitor always returns an error key which is set to an empty string if successful
                    capture_error = True
                    # check if a Kapacitor task with the given identifier already exists, if so, delete it and create it again
                    kapacitor_api_tasks_instance_url = "{0}/{1}".format(kapacitor_api_tasks_url, task_id)
                    if patch_duplicates and get(kapacitor_api_tasks_instance_url).status_code == 200:
                        # this will only happen if the patch_duplicates flag is set to True
                        delete(kapacitor_api_tasks_instance_url)
                        # an alternative is to use the PATCH API endpoint instead of deleting and creating it again, however,
                        # using the PATCH method requires the task to be disabled and re-enabled for changes to take effect (more HTTP requests)
                        response = post(kapacitor_api_tasks_url, json=kapacitor_http_request_body)
                        response_content = response.json()
                        capture_error = response_content.get("error", "") != ""

                    # if the error was not handled by patching an existing Kapacitor task, then add it in the list of errors
                    if capture_error:
                        alert_tasks_errors.append({
                            "policy": policy_id,
                            "trigger": event_id,
                            "error": response_content["error"]
                        })

                # extract http handlers
                http_handlers = trigger.trigger_tpl["action"]["implementation"]

                # subscribe all http handlers to the created topic
                self._config_kapacitor_alert_handlers(kapacitor_host, kapacitor_port, sfc, sfc_instance, policy_id, resource_spec_trigger_id, topic_id, event_id,
                                                      http_handlers, alert_handlers_errors, patch_duplicates=patch_duplicates)

        return alert_tasks_errors, alert_handlers_errors

    def _config_kapacitor_alert_handlers(self, kapacitor_host, kapacitor_port, sfc, sfc_i, policy_id, trigger_id,
                                         topic_id, event_id, http_handlers, alert_handlers_errors, patch_duplicates=False):
        """
        Handles the configuration of HTTP Post alert handlers.

        :param kapacitor_host: default host is localhost (CLMC service running on the same machine as Kapacitor)
        :param kapacitor_port: default value to use is 9092
        :param sfc: service function chain identifier
        :param sfc_i: service function chain instance identifier
        :param policy_id: policy ID those triggers relate to
        :param trigger_id: the resource specification trigger ID
        :param topic_id: topic ID built of sfc, sfc instance and event_id
        :param event_id: name of trigger
        :param http_handlers: list of handlers to subscribe
        :param alert_handlers_errors: the list for tracking errors while interacting with Kapacitor alert handlers
        :param patch_duplicates: (defaults to False) if set to True, any duplication errors will be handled by first deleting the existing resource and then creating it
        """

        kapacitor_api_handlers_url = "http://{0}:{1}{2}/{3}/handlers".format(kapacitor_host, kapacitor_port, self.KAPACITOR_TOPIC_API_PREFIX, topic_id)

        for http_handler_url in http_handlers:

            # check for flame_sfemc entry, if found replace with sfemc FQDN
            if http_handler_url == SFEMC:
                sfemc_fqdn = self.request.registry.settings['sfemc_fqdn']
                sfemc_port = self.request.registry.settings['sfemc_port']
                http_handler_url = "http://{0}:{1}/sfemc/event/{2}/{3}/{4}".format(sfemc_fqdn, sfemc_port, sfc, policy_id, trigger_id)

            handler_id = "{0}\n{1}\n{2}\n{3}\n{4}".format(sfc, sfc_i, policy_id, event_id, http_handler_url)
            handler_id = self.get_hash(handler_id)
            kapacitor_http_request_body = fill_http_post_handler_vars(handler_id, http_handler_url)
            response = post(kapacitor_api_handlers_url, json=kapacitor_http_request_body)
            response_content = response.json()
            log.info("{0} - {1}".format(response_content, response.status_code))

            if response_content.get("error", "") != "":
                capture_error = True
                # check if a Kapacitor handler with the given identifier already exists, if so, delete it and create it again
                kapacitor_api_handlers_instance_url = "{0}/{1}".format(kapacitor_api_handlers_url, handler_id)
                if patch_duplicates and get(kapacitor_api_handlers_instance_url).status_code == 200:
                    # this will only happen if the patch_duplicates flag is set to True
                    delete(kapacitor_api_handlers_instance_url)
                    response = post(kapacitor_api_handlers_url, json=kapacitor_http_request_body)
                    response_content = response.json()
                    capture_error = response_content.get("error", "") != ""

                # if the error was not handled by patching an existing Kapacitor handler, then add it in the list of errors
                if capture_error:
                    alert_handlers_errors.append({
                        "policy": policy_id,
                        "trigger": event_id,
                        "handler": http_handler_url,
                        "error": response_content["error"]
                    })

    def _parse_alert_spec(self, alert_spec_reference):
        """
        Parses an alert specification file to TOSCA template and validates it against the CLMC alerts specification schema.

        :param alert_spec_reference: the alert specification file received in the request

        :return: the parsed tosca template
        """

        # check that the alerts specification file was sent correctly
        if not hasattr(alert_spec_reference, "file") or not hasattr(alert_spec_reference, "filename"):
            raise HTTPBadRequest("Request to this API endpoint must include a (YAML) file input referenced as 'alert-spec' representing the TOSCA Alert Specification.")

        # extract alert specification file and filename
        alerts_input_filename = alert_spec_reference.filename
        alerts_input_file = alert_spec_reference.file

        # allow only .yaml and .yml extensions for filename
        if not alerts_input_filename.lower().endswith('.yaml') and not alerts_input_filename.lower().endswith('.yml'):
            raise HTTPBadRequest("Request to this API endpoint must include a (YAML) file input referenced as 'alert-spec' representing the TOSCA Alerts Specification.")

        # parse the alerts specification file
        try:
            alerts_yaml_content = load(alerts_input_file)
            adjust_tosca_definitions_import(alerts_yaml_content)
        except YAMLError as err:
            log.error("Couldn't parse user request file {0} to yaml format due to error: {1}".format(alerts_input_filename, err))
            log.error("Invalid content is: {0}".format(alerts_input_file.read()))
            raise HTTPBadRequest("Request alert specification file could not be parsed as valid YAML document.")

        # convert to tosca template
        try:
            tosca_tpl = ToscaTemplate(yaml_dict_tpl=alerts_yaml_content)
        except Exception as e:
            log.error(e)
            raise HTTPBadRequest("Request alert specification file could not be parsed as a valid TOSCA document.")

        # validate against CLMC spec.
        valid_alert_spec, error = validate_clmc_alerts_specification(tosca_tpl.tpl, include_error=True)
        if not valid_alert_spec:
            log.error("Error encountered while validating alerts specification against CLMC alerts specification schema - {0}".format(error))
            raise HTTPBadRequest("Request alert specification file could not be validated as a CLMC TOSCA alerts specification document.")

        return tosca_tpl

    def _parse_resource_spec(self, resource_spec_reference):
        """
        Parses a resource specification file to TOSCA template and extracts the information needed to compare it with the CLMC alerts specification.

        :param resource_spec_reference: the resource specification file received in the request

        :return: sfc ID, sfc instance ID, list of policy-trigger identifiers found in the resource spec.
        """

        # check that the resource specification file was sent
        if not hasattr(resource_spec_reference, "file") or not hasattr(resource_spec_reference, "filename"):
            raise HTTPBadRequest("Request to this API endpoint must include a (YAML) file input referenced as 'resource-spec' representing the TOSCA Resource Specification.")

        # extract the required information from the resource specification and return an error if an exception was encountered
        try:
            resource_spec_sfc, resource_spec_sfc_instance, resource_spec_policy_triggers = get_resource_spec_policy_triggers(resource_spec_reference)
        except Exception as e:
            log.error("Couldn't extract resource specification event IDs due to error: {0}".format(e))
            raise HTTPBadRequest("Couldn't extract resource specification event IDs - invalid TOSCA resource specification.")

        return resource_spec_sfc, resource_spec_sfc_instance, resource_spec_policy_triggers

    @staticmethod
    def get_hash(message):
        """
        Returns the hash value of a message encoded with utf-8 using hash algorithm sha256

        :param message: the message to hash

        :return: the value of the hash
        """

        byte_str = bytes(message, encoding="utf-8")
        hash_obj = sha256(byte_str)

        return hash_obj.hexdigest()
